// Copyright (C) 2007-2008  Gelenberg Andrej
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

#include <stdio.h>
#include <stdlib.h>
#include <libintl.h>
#include <string.h>
#include <locale.h>
#include <unistd.h>
#include <sys/stat.h>

#include "dbused.h"
#include "rule.h"
#include "bus.h"
#include "util.h"

const char * const prog_name = PROGRAMM_NAME;
char verbose = 0;

inline static void free_dfrules(struct df_rule_el *el)
{
	struct df_rule_el *prev;
	while ( el )
	{
		prev = el;
		el = el->next;
		free(prev->rule);
		free(prev);
	}
}

inline static void print_usage(const char *prog_name)
{
	fprintf(stderr,
		"%s: %s [-v] [-s] [-f] {%s}\n\t-v %s\n\t-s %s\n\t-f %s\n",
		gettext("usage"),
		prog_name,
		gettext("match rules file"),
		gettext("verbose mode"),
		gettext("syntax check only"),
		gettext("fork and become daemon")
		);
}

int main(int argc, char *argv[])
{
	enum error retval;
	char not_syntax_only = 1;
	char to_fork = 0;

	setlocale(LC_ALL, "");
	bindtextdomain(prog_name, NULL);
	textdomain(prog_name);

	char **arg = &argv[1];

	if ( *arg && (!strcmp("-v", *arg)) )
	{
		verbose = 1;
		arg++;
		printf("* %s\n", gettext("verbose mode is on"));
	}

	if ( *arg && (!strcmp("-s", *arg)) )
	{
		not_syntax_only = 0;
		arg++;
		if ( verbose )
		{
			printf("* %s\n", gettext("check syntax only"));
		}
	}

	if ( *arg && (!strcmp("-f", *arg)) )
	{
		to_fork = 1;
		arg++;
		if ( verbose )
		{
			printf("* %s\n", gettext("fork and become daemon"));
		}
	}


	if ( verbose )
	{
		printf("DED (DBus Event Daemon)\nAndrej Gelenberg 2007-2008(c)\n");
	}

	struct dbus_init_data did;
	struct err_s error;

	if ( verbose )
	{
		printf("* %s: %s\n", gettext("load file"), *arg);
	}

	if ( !*arg ) // Prüfung von Argumentenzahl
	{
		fprintf(stderr, "%s\n", gettext("Too few arguments"));
		print_usage(argv[0]);
		return ERR_EXIT_ARGS;
	}


	if ( (retval = load_rules(*arg, &did, &error)) ) // Laden von Matchregeln
	{
		fprintf(stderr, "%s \"%s\": %s\n",
			gettext("Error by loading of rules file"),
			*arg, error.msg);
		err_free(error);
		return retval;
	}

	if ( !did.mr_root )
	{
		fprintf(stderr, "%s\n", gettext("No match rool tree was found."));
		free_dfrules(did.df_rule_start);
		free_match_tree(did.mr_root);
		return ERR_EXIT_NO_MRTREE_FOUND;
	}

	if ( not_syntax_only )
	{
		if ( to_fork )
		{
			if ( verbose )
			{
				verbose = 0;
				printf("* %s\n", gettext("Fork and become daemon. Verbose mode is turn off."));
			}

			pid_t p = fork();
			if ( p < 0 )
			{
				fprintf(stderr, "%s\n", gettext("Can't fork"));
				return ERR_EXIT_FORK;
			}

			if ( p > 0 )
			{
				return NO_ERROR;
			}
			umask(0);
			p = setsid();
			if ( p < 0 )
			{
				return ERR_EXIT_SID;
			}
			if ( chdir("/") < 0 )
			{
				return ERR_EXIT_CHDIR;
			}
			
			close(STDIN_FILENO);
			close(STDERR_FILENO);
			close(STDOUT_FILENO);
		}

		if ( (retval = start_dbus_monitor(&error, &did)) )
		{
			fprintf(stderr, "%s: %s\n", gettext("Error by dbus monitor start"), error.msg);
			err_free(error);
			free_dfrules(did.df_rule_start);
			free_match_tree(did.mr_root);
			return retval;
		}
	}

	free_dfrules(did.df_rule_start);
	free_match_tree(did.mr_root);

	return NO_ERROR;
}
