// Copyright (C) 2007-2008  Gelenberg Andrej
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

#include <stdlib.h>
#include <stdio.h>
#include <libintl.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>

#include "match.h"
#include "dbused.h"
#include "util.h"

static const char true_str[] = "true";
static const char false_str[] = "false";

void free_match_tree(struct match_rule *root)
{
	if ( root )
	{
		free_match_tree(root->next);
		free_match_tree(root->fork);
		enum match_rule_type mr_type = root->type;
		if ( (mr_type != MESSAGE_TYPE) && (mr_type != NONE) )
		{
			switch ( mr_type )
			{
				case ARG:
				{
					struct arg_data *d = root->data.arg->data;
					enum arg_type t = d->type;
					if ( t == STRING )
					{
						free(d->data.str);
					}
					else if ( t == REGEX )
					{
						regex_t *r = d->data.regex;
						regfree(r);
						free(r);
					}
					free(d);
					break;
				}
				case SEND:
				{
					dbus_message_unref(root->data.send->msg);
				}
			}
			free(root->data.ptr);
		}
		free(root);
	}
}

static void action_send(struct send_t *s)
{
	DBusConnection *conn;
	DBusError error;
	dbus_error_init(&error);
	conn = dbus_bus_get(s->bus, &error);
	if ( conn == NULL )
	{
		fprintf(stderr, "%s: %s\n", gettext("Error by getting bus"), error.message);
		dbus_error_free(&error);
		return;
	}
	if ( verbose )
	{
		DBusMessage *reply;
		reply = dbus_connection_send_with_reply_and_block(conn,
								s->msg, -1,
								&error);
		if ( dbus_error_is_set (&error) )
		{
			printf("* %s: %s: %s\n", gettext("Error by send action"),
				error.name,
				error.message);
			dbus_error_free(&error);
		}
		else if ( reply )
		{
			dbus_message_unref(reply);
		}
	}
	else
	{
		dbus_connection_send(conn, s->msg, 0);
	}
}

inline static char match(DBusMessage *msg, struct match_rule *mr_root)
{
	switch ( mr_root->type )
	{
		case NONE:
			fprintf(stderr, "%s\n", gettext("Match warning: NONE match rule was found."));
			return 1;
		case PATH:
			return dbus_message_has_path(msg, mr_root->data.path);
		case INTERFACE:
			return dbus_message_has_interface(msg, mr_root->data.interface);
		case MEMBER:
			return dbus_message_has_member(msg, mr_root->data.member);
		case DESTINATION:
			return dbus_message_has_destination(msg, mr_root->data.destination);
		case SENDER:
			return dbus_message_has_sender(msg, mr_root->data.sender);
		case MESSAGE_TYPE:
			return (dbus_message_get_type(msg) == mr_root->data.message_type);
		case ARG:
		{
			struct arg_match *arg = mr_root->data.arg;
			DBusMessageIter iter;
			if (!dbus_message_iter_init(msg, &iter))
			{
				return 0;
			}
			unsigned char i;
			unsigned char n = arg->num;
			for ( i=0; i < n; ++i)
			{
				if (!dbus_message_iter_next(&iter))
				{
					return 0;
				}
			}
			int message_type = dbus_message_iter_get_arg_type(&iter);
			struct arg_data *adata = arg->data;
			enum arg_type atype  = adata->type;
			switch (atype)
			{
				case STRING:
				{
					if ( message_type != DBUS_TYPE_STRING )
					{
						return 0;
					}
					char *str;
					dbus_message_iter_get_basic(&iter, &str);
					return ( ! strcmp(str, adata->data.str) );
				}
				case REGEX:
				{
					if ( message_type != DBUS_TYPE_STRING )
					{
						return 0;
					}
						char *str;
						dbus_message_iter_get_basic(&iter, &str);
						return ( ! regexec(adata->data.regex, str, 0, 0, 0) );
				}
			}
			if ( atype != message_type )
			{
				return 0;
			}
			switch (atype)
			{
				case BYTE:
				case BOOL:
				{
					char c;
					dbus_message_iter_get_basic(&iter, &c);
					return ( c == adata->data.byte_d );
				}
				case INT16:
				case UINT16:
				{
					dbus_int16_t i;
					dbus_message_iter_get_basic(&iter, &i);
					return ( i == adata->data.int16_d );
				}
				case INT32:
				case UINT32:
				{
					dbus_int32_t i;
					dbus_message_iter_get_basic(&iter, &i);
					return ( i == adata->data.int32_d );
				}
				case INT64:
				case UINT64:
				{
					dbus_int64_t i;
					dbus_message_iter_get_basic(&iter, &i);
					return ( i == adata->data.int64_d );
				}
				case DOUBLE:
				{
					double d;
					dbus_message_iter_get_basic(&iter, &d);
					return ( d == adata->data.double_d );
				}
			}
		}
		case EXEC:
		{
			if ( verbose )
			{
				printf("* %s: %s\n", gettext("action exec"), mr_root->data.cmd);
			}
			int c;
			if ( (c = system(mr_root->data.cmd)) && verbose )
			{
				printf("\t%s: %d\n", gettext("error code"), c);
			}
			return 0;
		}
		case FORK:
		{
			if ( verbose )
			{
				printf("* %s: %s\n", gettext("action fork"), mr_root->data.cmd);
			}
			pthread_t thread;
			pthread_create(&thread, 0, system, mr_root->data.cmd);
			pthread_detach(thread);
			return 0;
		}
		case SEND:
		{
			if ( verbose )
			{
				printf("* %s\n", gettext("action send"));
			}
			pthread_t thread;
			pthread_create(&thread, 0, action_send, mr_root->data.ptr);
			pthread_detach(thread);
			return 0;
		}
		default:
			fprintf(stderr, "%s\n", gettext("Match error: unknown match type was found."));
		}
	return 0;
}

static void matching(DBusMessage *msg, struct match_rule *mr_root)
{
	struct match_rule *next;
	if ( match(msg, mr_root) )
	{
		next = mr_root->next;
		if ( next )
		{
			matching(msg, next);
		}
		else
		{
			fprintf(stderr, "%s\n", gettext("Matching error: next element are waited, but no found."));
		}
	}
	next = mr_root->fork;
	if ( next )
	{
		matching(msg, next);
	}
}

DBusHandlerResult match_rule_filter(DBusConnection *connection, DBusMessage *msg, struct match_rule *mr_root)
{
	matching(msg, mr_root);
	return DBUS_HANDLER_RESULT_HANDLED;
}
